#include <config.h>
module wavefunction_class
  use comlib
  use lattice_class
  use field_gauge_class
  use field_fermion_class
  implicit none
  private

  public :: TYPE_WF_LOCAL
  public :: TYPE_WF_JACOBI

  public :: wavefunc
  public :: new
  public :: delete
  public :: print
  public :: read
  public :: set_source

  integer, parameter :: TYPE_WF_LOCAL  = 0
  integer, parameter :: TYPE_WF_JACOBI = 1

  type :: wavefunc
    integer :: type = TYPE_WF_LOCAL
    integer :: itx0 = 1
    integer :: ity0 = 1
    integer :: itz0 = 1
    integer :: itt0 = 1
    integer ::  Nsmr = 1
    real(DP) :: Ksmr = 0.1_DP
  end type

  interface new
    module procedure new_wf
  end interface

  interface delete
    module procedure delete_wf
  end interface

  interface read
    module procedure read_wf
  end interface

  interface print
    module procedure print_wf
  end interface

  interface set_source
    module procedure set_source_wf
  end interface

contains

subroutine new_wf(this)
  implicit none
  type(wavefunc), intent(inout) :: this
  this%type = TYPE_WF_LOCAL
  this%itx0 = 1
  this%ity0 = 1
  this%itz0 = 1
  this%itt0 = 1
  this%Nsmr = 1
  this%Ksmr = 0.1_DP
  return
end subroutine

subroutine delete_wf(this)
  implicit none
  type(wavefunc), intent(inout) :: this
  call new(this)
  return
end subroutine

subroutine read_wf(this,iout)
  use error_class
  implicit none
  type(wavefunc), intent(inout) :: this
  integer, intent(in) :: iout
  integer :: flg
  flg = 0
  if (0==nodeid) then
    read(iout,*) this%type
    read(iout,*) this%itx0,this%ity0,this%itz0,this%itt0
  endif

#ifndef _singlePU
  call comlib_bcast(this%type,0)
  call comlib_bcast(this%itx0,0)
  call comlib_bcast(this%ity0,0)
  call comlib_bcast(this%itz0,0)
  call comlib_bcast(this%itt0,0)
#endif

  select case(this%type)
  case (TYPE_WF_LOCAL)
    continue
  case (TYPE_WF_JACOBI)
    if (0==nodeid) then
      read(iout,*)this%Nsmr,this%Ksmr
    endif
#ifndef _singlePU
    call comlib_bcast(this%Nsmr,0)
    call comlib_bcast(this%Ksmr,0)
#endif
  case default
    call error_stop("Error in wavefunction type parameter.")
  end select

  if (this%itx0 <= 0 .or. this%itx0 > NTX) then
    call error_stop("Error in wavefunction itx0 parameter.")
  endif
  if (this%ity0 <= 0 .or. this%ity0 > NTY) then
    call error_stop("Error in wavefunction ity0 parameter.")
  endif
  if (this%itz0 <= 0 .or. this%itz0 > NTZ) then
    call error_stop("Error in wavefunction itz0 parameter.")
  endif
  if (this%itt0 <= 0 .or. this%itt0 > NTT) then
    call error_stop("Error in wavefunction itt0 parameter.")
  endif

  select case(this%type)
  case (TYPE_WF_LOCAL)
    continue
  case (TYPE_WF_JACOBI)
    if (this%Nsmr < 0) then
      call error_stop("Error in wavefunction Nsmr parameter.")
    endif
    if (this%Ksmr < 0) then
      call error_stop("Error in wavefunction Ksmr parameter.")
    endif
  end select

  return
end subroutine

subroutine print_wf(this)
  use error_class
  implicit none
  type(wavefunc), intent(in) :: this
  if (0==nodeid) then
    select case(this%type)
    case(TYPE_WF_LOCAL)
      write(*,'(" wave function type: LOCAL")')
    case(TYPE_WF_JACOBI)
      write(*,'(" wave function type: JACOBI")')
    end select
    write(*,'(" wave function centor location: (x,y,z,t)=(",I3,",",I3,",",I3,",",I3,")")') &
 &   this%itx0,this%ity0,this%itz0,this%itt0
  endif
  return
end subroutine

subroutine set_source_wf(this,jc,js,y,u)
  use error_class
  implicit none
  type(wavefunc), intent(in) :: this
  integer, intent(in) :: jc,js
  type(field_quark_wg),  intent(inout) :: y
  type(vfield_gluon_wg), intent(in)    :: u
  character(CHARLEN) :: str
  integer :: ix,iy,iz,it,is,ic
  integer :: ieo,ieoxyz,ith
  integer :: ix0,iy0,iz0
  integer :: ipx0,ipy0,ipz0

  !
  ! source center corrdinate in a node
  !
  ix0 = mod(this%itx0-1,NX)+1
  iy0 = mod(this%ity0-1,NY)+1
  iz0 = mod(this%itz0-1,NZ)+1

  !
  ! source center node corrdinate
  !
  ipx0 = (this%itx0-1)/NX
  ipy0 = (this%ity0-1)/NY
  ipz0 = (this%itz0-1)/NZ

  call new(y)
  call clear(y)

  !
  ! set local source
  !
  if ( (ipsite(1) == ipx0) .and. &
 &     (ipsite(2) == ipy0) .and. &
 &     (ipsite(3) == ipz0) ) then
    ix = ix0
    iy = iy0
    iz = iz0
    it = this%itt0
    ieo = mod(ipeo+it+iz+iy+ix,2)
    ith = it/2
    y%eo(ieo)%s(ith,iz,iy,ix)%y(jc,js) = Z1
  endif

  select case(this%type)
  case (TYPE_WF_LOCAL)
    continue
  case (TYPE_WF_JACOBI)
    call jacobi_smear(this,y,u)
  case default
    write(str,'("error in type wavefunc:",I4)')this%type
    call error_stop(TRIM(str))
  end select

  return
end subroutine

subroutine jacobi_smear(this,fy,u)
  implicit none
  type(wavefunc), intent(in) :: this
  type(field_quark_wg),  intent(inout) :: fy
  type(vfield_gluon_wg), intent(in)    :: u
  type(field_quark_wg), allocatable :: y,w

  type(su3fv_spinor) :: yt
  real(DP) :: rnorm
  integer :: ix,iy,iz,it,is,ic
  integer :: ieo,ith,ioe
  integer :: iter

  allocate(y,w)
  call new(y)
  call new(w)
  call clear(w)
  call assign(y,fy)
  
  it = this%itt0

  do iter=1,this%Nsmr

    call copy_boundary(fy)

    !
    !  fy = y + kappa M fy
    !
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieo,ioe,ith,ic,is,yt)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ

      ith = it/2
      ieo = mod(ipeo+it+iz+iy+ix,2)
      ioe = 1-ieo

      yt%y(:,:) = Z0

      !
      ! Z-direction
      !
      do is=1,SPIN
      do ic=1,COL
        yt%y(ic,is) = yt%y(ic,is) + u%eo(ieo)%mu(3)%s(ith,iz  ,iy,ix)%u(ic,1)  &
  &                                     *fy%eo(ioe)%s(ith,iz+1,iy,ix)%y(1,is)  &
  &                               + u%eo(ieo)%mu(3)%s(ith,iz  ,iy,ix)%u(ic,2)  &
  &                                     *fy%eo(ioe)%s(ith,iz+1,iy,ix)%y(2,is)  &
  &                               + u%eo(ieo)%mu(3)%s(ith,iz  ,iy,ix)%u(ic,3)  &
  &                                     *fy%eo(ioe)%s(ith,iz+1,iy,ix)%y(3,is)  &
  &                         + conjg(u%eo(ioe)%mu(3)%s(ith,iz-1,iy,ix)%u(1,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz-1,iy,ix)%y(1,is)  &
  &                         + conjg(u%eo(ioe)%mu(3)%s(ith,iz-1,iy,ix)%u(2,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz-1,iy,ix)%y(2,is)  &
  &                         + conjg(u%eo(ioe)%mu(3)%s(ith,iz-1,iy,ix)%u(3,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz-1,iy,ix)%y(3,is)
      enddo
      enddo
    
      !
      ! Y-direction
      !
      do is=1,SPIN
      do ic=1,COL
        yt%y(ic,is) = yt%y(ic,is) + u%eo(ieo)%mu(2)%s(ith,iz,iy  ,ix)%u(ic,1)  &
  &                                     *fy%eo(ioe)%s(ith,iz,iy+1,ix)%y(1,is)  &
  &                               + u%eo(ieo)%mu(2)%s(ith,iz,iy  ,ix)%u(ic,2)  &
  &                                     *fy%eo(ioe)%s(ith,iz,iy+1,ix)%y(2,is)  &
  &                               + u%eo(ieo)%mu(2)%s(ith,iz,iy  ,ix)%u(ic,3)  &
  &                                     *fy%eo(ioe)%s(ith,iz,iy+1,ix)%y(3,is)  &
  &                         + conjg(u%eo(ioe)%mu(2)%s(ith,iz,iy-1,ix)%u(1,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz,iy-1,ix)%y(1,is)  &
  &                         + conjg(u%eo(ioe)%mu(2)%s(ith,iz,iy-1,ix)%u(2,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz,iy-1,ix)%y(2,is)  &
  &                         + conjg(u%eo(ioe)%mu(2)%s(ith,iz,iy-1,ix)%u(3,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz,iy-1,ix)%y(3,is)
      enddo
      enddo

      !
      ! X-direction
      !
      do is=1,SPIN
      do ic=1,COL
        yt%y(ic,is) = yt%y(ic,is) + u%eo(ieo)%mu(1)%s(ith,iz,iy,ix  )%u(ic,1)  &
  &                                     *fy%eo(ioe)%s(ith,iz,iy,ix+1)%y(1,is)  &
  &                               + u%eo(ieo)%mu(1)%s(ith,iz,iy,ix  )%u(ic,2)  &
  &                                     *fy%eo(ioe)%s(ith,iz,iy,ix+1)%y(2,is)  &
  &                               + u%eo(ieo)%mu(1)%s(ith,iz,iy,ix  )%u(ic,3)  &
  &                                     *fy%eo(ioe)%s(ith,iz,iy,ix+1)%y(3,is)  &
  &                         + conjg(u%eo(ioe)%mu(1)%s(ith,iz,iy,ix-1)%u(1,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz,iy,ix-1)%y(1,is)  &
  &                         + conjg(u%eo(ioe)%mu(1)%s(ith,iz,iy,ix-1)%u(2,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz,iy,ix-1)%y(2,is)  &
  &                         + conjg(u%eo(ioe)%mu(1)%s(ith,iz,iy,ix-1)%u(3,ic)) &
  &                                     *fy%eo(ioe)%s(ith,iz,iy,ix-1)%y(3,is)
      enddo
      enddo

      w%eo(ieo)%s(ith,iz,iy,ix)%y(:,:) = y%eo(ieo)%s(ith,iz,iy,ix)%y(:,:) &
 &                                                  + this%Ksmr*yt%y(:,:)
    enddo
    enddo
    enddo

    call assign(fy,w)

  enddo

  deallocate(y,w)

  !
  ! normalize wavefunction
  !
  rnorm = 0.0_DP
  it = this%itt0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieo,ith,ic,is) REDUCTION(+:rnorm)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ith = it/2
    ieo = mod(ipeo+it+iz+iy+ix,2)
    do is=1,SPIN
    do ic=1,COL
      rnorm = rnorm +  real(fy%eo(ieo)%s(ith,iz,iy,ix)%y(ic,is))**2 &
 &                  + aimag(fy%eo(ieo)%s(ith,iz,iy,ix)%y(ic,is))**2
    enddo
    enddo
  enddo
  enddo
  enddo
#ifndef _singlePU
  call comlib_sumcast(rnorm)
#endif
  rnorm = 1.0_DP/sqrt(rnorm)

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieo,ith)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ith = it/2
    ieo = mod(ipeo+it+iz+iy+ix,2)
    fy%eo(ieo)%s(ith,iz,iy,ix)%y(:,:) = fy%eo(ieo)%s(ith,iz,iy,ix)%y(:,:)*rnorm
  enddo
  enddo
  enddo

  return
end subroutine

end module
